----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    23:06:54 02/03/2010 
-- Design Name: 
-- Module Name:    PET816 - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 1.0A - First revision
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity PET816b is
	Port( 			
			fastclkin : in STD_LOGIC;		-- 50MHz in
			
			-- PET interface

			phi0 : in STD_LOGIC;			-- PET Phi1 in
			rdy : in STD_LOGIC;
			reset : in STD_LOGIC;
			
			norom : out STD_LOGIC;		-- switch off PET ROM when set to '0'
			rnw : out STD_LOGIC;
			
			-- 65816 interface

			cpuaddrin_64k : in STD_LOGIC_VECTOR (23 downto 16);
			cpuaddrin_2k : in STD_LOGIC_VECTOR (15 downto 11);
			cpuaddr_ismyio : in STD_LOGIC;		-- when 0 then addr bits 4 to 11 are all 1
			cpuaddrin_16 : in STD_LOGIC_VECTOR (3 downto 0);
			cpudata : inout STD_LOGIC_VECTOR (7 downto 0);
			
			cpuvpa : in STD_LOGIC;
			cpuvda : in STD_LOGIC;
			cpuvp : in STD_LOGIC;
			cpurnw : in STD_LOGIC;
			
			cpuclk : out STD_LOGIC;
			
			-- control
			
			nslowbusen : out STD_LOGIC;			-- when 0, pass CPU address bus to PET
			nslowdataen : out STD_LOGIC;			-- when 0, pass CPU data bus to PET
			latchen: out STD_LOGIC;					-- '573 latch control (1=passthrough, 0=latch)
			
			-- config
			bootromin: in STD_LOGIC;				-- when 1 then jumper is set to enable boot rom
			
			-- fast memory interface (2 RAM modules)
			
			nramcs : out STD_LOGIC_VECTOR (1 downto 0);
			
			-- bootrom select, in-system-programming
			nromcs : out STD_LOGIC;
			nromwe : out STD_LOGIC;
			
			-- external logic:
			-- phi1 <= not(phi0);
			-- phi2 <= not(phi1);
			-- nramoe <= not(cpurnw);
			-- cpuaddr_ismyio <= not( cpuaddrin(11 downto 4) = "1111111" );
		
			diag : out STD_LOGIC;

			diag_fastclkby2 : out std_logic;
			diag_fastclkbydiv : out std_logic;
			diag_slowdetected : out std_logic;
			diag_slowaccess : out std_logic;
			diag_slowdone : out std_logic;
			diag_startfast : out std_logic;
			diag_fakeclk : out std_logic;
			diag_extclkdel : out std_logic;
			diag_clklatch : out std_logic_vector(3 downto 0);
			diag_clkcounter : out std_logic_vector(3 downto 0);
			diag_validaddr : out std_logic
	);
			
end PET816b;

architecture Behavioral of PET816b is

		-- internal state signals
		signal clkdiv: std_logic_vector (1 downto 0);	-- clock divisor index
		signal slowonly: std_logic := '1';					-- when 1 then only slow access is allowed
		signal slow64k: std_logic := '1';					-- when 1 low 64k are mapped to PET (otherwise I/O, video only)
																		--	writes are passed through to fast RAM (with slow access)
		signal bootrom: std_logic := '0';					-- when 1, map expansion ROM to low 64k; initialied to jumper
		signal prgrom: std_logic := '0';						-- when 1, allow writes to ROM (in system programming)
		signal wprotect: std_logic_vector (1 downto 0);	-- when 1, write protect fast RAM map
		
		-- master timing
		signal fastclkbydiv: std_logic := '0';					-- fast clock divided by clkcount
		signal fastclkby2: std_logic := '0';				-- fastclkin divided by 2
		signal clkcounter: std_logic_vector (3 downto 0);
	
		signal clklatch: std_logic_vector(3 downto 0);

		-- clock shift register
		signal phi0D1: std_logic := '0';
		signal phi0D2: std_logic := '0';
		signal phi0D3: std_logic := '0';
		signal phi0D4: std_logic := '0';
		signal phi0D5: std_logic := '0';
		signal phi0D6: std_logic := '0';
		signal phi0D7: std_logic := '0';
		signal extclkdel: std_logic := '0';	-- final delayed value

		-- bus timing	
--		signal q3: std_logic := '0';
		-- state engine
		signal slowdetected: std_logic := '0';
		signal slowaccess: std_logic := '0';
		signal slowdone: std_logic := '0';
		signal startfast: std_logic := '0';
		signal fakeclk: std_logic := '0';					-- clk signal between slowdetected and actual slow access
		
		signal cpuclkint: std_logic := '0';
		
		signal validaddr: std_logic := '0';					-- TODO: vpa/vda/vp analysis
		
		signal isio: std_logic := '0';			-- 1 when the I/O area is selected (valid together with islow64k)
		signal isvideo: std_logic := '0';		-- 1 when the video area is selected (valid together with islow64k)
		signal islow64k: std_logic := '0';		-- 1 when the lowest 64k are selected
		signal isorig: std_logic := '0';			-- 1 when the original system is being addressed
		
		signal myregister: std_logic := '0';	-- 1 when accesing own register, to switch off data bus driver
begin
		-----------------------------------------------------------------------------------------------
		-- diag

		diag_fastclkby2 <= fastclkby2;
		diag_fastclkbydiv <= fastclkbydiv;
		diag_slowdetected <= slowdetected;
		diag_slowaccess <= slowaccess;
		diag_slowdone <= slowdone;
		diag_startfast <= startfast;
		diag_fakeclk <= fakeclk;

		diag_extclkdel <= extclkdel;

		diag_clkcounter <= clkcounter;
		diag_clklatch <= clklatch;

		diag_validaddr <= validaddr;

		-----------------------------------------------------------------------------------------------
		-- system specific part

		DoPetIO: process(cpuaddrin_2k, cpuaddrin_64k, slow64k, bootrom)
		begin
			if (cpuaddrin_2k = "11101") then 					-- $00e800-$00efff
				isio <= '1';
			else
				isio <= '0';
			end if;

			if (cpuaddrin_2k(15 downto 12) = "1000") then 	-- $008xxx
				isvideo <= '1';
			else
				isvideo <= '0';
			end if;

			if (cpuaddrin_64k = "00000000") then
				islow64k <= '1';
			else
				islow64k <= '0';
			end if;
			
			-- use orig system always for video and I/O, plus when config is orig64k and not boot rom
			if (islow64k = '1'
					and (isio = '1'
						or isvideo = '1'
						or (slow64k = '1'
							and bootrom = '0')
					)
				) then
				isorig <= '1';
			else
				isorig <= '0';
			end if;
		end process DoPetIO;
		
		-----------------------------------------------------------------------------------------------
		-- master timing 
		Timing: process(fastclkin)
		begin
			if (falling_edge(fastclkin)) then
				fastclkby2 <= not(fastclkby2);
			end if;
		end process Timing;

		-- master timing 
		Master_Timing: process(fastclkin, reset, slowaccess)
		begin
			if (reset = '0') then
			        clkcounter <= "0000";
				fastclkbydiv <= '0';
			elsif (slowaccess = '1') then
				clkcounter <= clklatch;
				fastclkbydiv <= '0';
			elsif (falling_edge(fastclkin)) then
				
				if (clkcounter = 0) then
					clkcounter <= clklatch;
					fastclkbydiv <= not(fastclkbydiv);
				else
					clkcounter <= clkcounter - 1;
				end if;
			end if;
		end process Master_Timing;
		
		
		-----------------------------------------------------------------------------------------------
		-- bus timing generation
		-- pass phi1 through shift register by fastclock (50MHz)
		-- @50MHz -> @25MHz fastclkb2, each step is 40ns
		Delay1: process(fastclkby2)
		begin
			if (falling_edge(fastclkby2)) then
				phi0D1 <= phi0;
				phi0D2 <= phi0D1;
				phi0D3 <= phi0D2;
				phi0D4 <= phi0D3;
				phi0D5 <= phi0D4;
				phi0D6 <= phi0D5;
				phi0D7 <= phi0D6;
			end if;
		end process Delay1;

		extclkdel <= phi0D7;

		-----------------------------------------------------------------------------------------------
		-- cpu bus timing

		-- slow clock handling
		DoSlowAccess: process(extclkdel, startfast, slowdetected)
		begin
			if (startfast = '1') then
				slowaccess <= '0';
			elsif (falling_edge(extclkdel)) then
				slowaccess <= slowdetected;
			end if;
		end process DoSlowAccess;

		-- slow memory: orig mem and slow access (here: all negative logic)
		nslowbusen <= not(isorig) or not(slowaccess);
		-- slow memory; same as nslowbusen, but disable driver when reading own register
		nslowdataen <= not(isorig) or not(slowaccess) or myregister;
		-- slow memory r/-w
		rnw <= cpurnw or not (isorig) or not(slowaccess);
		-- switch off original ROMs when 0, 
		norom <= isorig;
		
		DoSlowDone: process(phi0, startfast, slowaccess)
		begin
			if (startfast = '1') then
				slowdone <= '0';
			elsif (falling_edge(phi0)) then
				-- wait for slow access to finish
				if (rdy = '1') then
					slowdone <= slowaccess;
				end if;
			end if;
		end process DoSlowDone;

		DoStartFast: process(phi0D2, phi0D3, slowaccess, phi0, slowdone)
		begin
			if (phi0D3 = '1' and phi0D2 = '0') then
				startfast <= '1'; 
			elsif (phi0D3 = '0') then
				startfast <= '0';
			end if;
		end process DoStartFast;

		DoFakeClock: process(slowdetected, extclkdel, slowaccess, reset)
		begin
			if (reset='0' or (extclkdel = '1' and slowaccess = '1')) then
				fakeclk <= '0';
			elsif (rising_edge(slowdetected)) then
				fakeclk <= '1';
			end if;
		end process DoFakeClock;
		
		-- CPU clock, syntesised
		cpuclkint <= 
				(fastclkbydiv and not(slowaccess or slowdetected))		-- fast access
				or (fakeclk)
				or ((phi0 or not(cpurnw)) and slowaccess); -- slow access, the rnw makes writes a bit longer
		
		cpuclk <= cpuclkint;
		latchen <= not(cpuclkint);

		diag <= fakeclk;

--		validaddr <= cpuvpa or cpuvda; -- or not(cpuvp);
		valid: process(fastclkbydiv) 
		begin
			if (reset='0') then 
				validaddr <= '0';
			elsif (falling_edge(fastclkbydiv)) then
				validaddr <= not(validaddr);
			end if;
		end process valid;	
				
		DoSlowDetect: process(cpuclkint, reset, validaddr, slowdone)
		begin
			if (reset = '0' or slowdone = '1') then
				-- reset
				slowdetected <= '0';
			elsif rising_edge(cpuclkint) then
				if (validaddr = '0' or slowdone = '1') then 
					-- pass quickly over invalid accesses
					-- do not spuriously detect directly after previous slow (slowdone=1)
					slowdetected <='0';
				elsif (slowonly = '1') then
					slowdetected <= '1';
				else
					if ( cpuaddrin_64k(23 downto 19) = "11111" ) then
						-- uppermost 512k
						slowdetected <= '1';
					elsif (islow64k = '1') then
						-- lowest 64k
						if (isio = '1'
							or isvideo = '1'
							or slow64k = '1' 
							or bootrom = '1'
							) then
							slowdetected <= '1';
						else
							-- fall through lowest 64k when not I/O, video, bootrom or slow64k setting
							slowdetected <= '0';
						end if;
					else
						-- all other memory above 64k except uppermost 512k 
						slowdetected <= '0';
					end if;
				end if;
			end if;
		end process DoSlowDetect;

		-----------------------------------------------------------------------------------------------
		-- CPU access to config registers

		DoWriteAccess: process(cpuclkint, reset, bootromin)
		begin
			if (reset = '0') then
				clkdiv <= "11";
				slowonly <= '1';
				slow64k <= '1';
				bootrom <= bootromin;
				prgrom <= '0';
				wprotect <= "00";
				clklatch <= "1010";
			elsif (falling_edge(cpuclkint) and cpurnw = '0') then
				if (islow64k = '1' 
						and isio = '1'
						and cpuaddr_ismyio = '0') then
					
					 case cpuaddrin_16 is 
						when "1111" =>
							-- write $00efff
							clkdiv <= cpudata(1 downto 0);
							wprotect <= cpudata(3 downto 2);
							slowonly <= cpudata(4);
							slow64k <= cpudata(5);
							bootrom <= cpudata(6);
							prgrom <= cpudata(7);
						when others =>
					end case;
					case clkdiv(1 downto 0) is
						when "00" =>
							clklatch <= "0001";	-- 1 = 12,5 MHz
						when "01" =>
							clklatch <= "0010";	-- 2 = 8,3 MHz
						when "10" => 
							clklatch <= "0101";	-- 5 = 4,2 MHz
						when "11" =>
							clklatch <= "1010";	-- 10 = 2,3 MHz
						when others =>
							clklatch <= "1111";
					end case;
				end if;
			end if;
		end process DoWriteAccess;

--		DoReadAccess: process(cpuclkint, phi0D2, reset)
--		begin
--			if (cpuclkint = '0') then 
--				cpudata <= "ZZZZZZZZ";
--				myregister <= '0';
--			elsif (rising_edge(phi0D2) and cpurnw = '1') then
--				if (islow64k = '1'
--						and isio = '1'
--						and cpuaddr_ismyio = '0') then
--					 case cpuaddrin_16 is
--						when "1111" =>
--							-- read $00efff
--							myregister <= '1';
--							cpudata <= (prgrom, bootrom, slow64k, slowonly, wprotect(1), wprotect(0), clkdiv(1), clkdiv(0));
--						when others =>
--							cpudata <= "ZZZZZZZZ";
--							myregister <= '0';
--					end case;
--				end if;
--			end if;
--		end process DoReadAccess;
		
		-----------------------------------------------------------------------------------------------
		-- memory select lines

		-- ram0 is always selected on write as well, but on read only where not slowdetected
		-- each one has 512k
		MemSel: process(cpuclkint, cpuaddrin_64k, cpurnw, reset, validaddr, 
						slowdetected, prgrom, bootrom, cpuaddrin_2k, isio,
						isorig, islow64k, isvideo, wprotect)
		begin
			if (reset = '0' or validaddr = '0' or cpuclkint = '0') then
				nramcs <= "11";
				nromcs <= '1';
				nromwe <= '1';
			else
				-- bank 0 ram, 0-512k
				if (cpuaddrin_64k(23 downto 19) = "00000")
					then
					if islow64k = '1' then
						if (cpurnw = '1' and (bootrom = '1' or isorig = '1')) then
							-- if read, but bootrom or isorig enabled, then not selected
							nramcs(0) <= '1';
						else
							-- write, or read and not bootrom and not isorig
							-- check write protect in low 64k
							case wprotect is
								when "00" =>
									nramcs(0) <= '0';		-- no write protect
								when "01" =>
									if (cpuaddrin_2k(15) = '1'
										and (cpuaddrin_2k(14) = '1'
											or cpuaddrin_2k(13) = '1')) then
										nramcs(0) <= not(cpurnw);		-- write protected $a000-$ffff
									else
										nramcs(0) <= '0';	-- ok
									end if;
								when "10" =>
									if (cpuaddrin_2k(15) = '1'
										and cpuaddrin_2k(14) = '1') then
										nramcs(0) <= not(cpurnw);		-- write protected $c000-$ffff
									else
										nramcs(0) <= '0';	-- ok
									end if;
								when "11" =>
									if (cpuaddrin_2k(15) = '1'
										and cpuaddrin_2k(14) = '1'
										and cpuaddrin_2k(13) = '1') then
										nramcs(0) <= not(cpurnw);		-- write protected $e000-$ffff
									else
										nramcs(0) <= '0';	-- ok
									end if;
								when others =>
									nramcs(0) <= '1';						-- no access, should not happen
							end case;
						end if;  -- read/bootrom/orig
					else	-- if islow64k
						-- 64k-512k
						nramcs(0) <= '0';
					end if;
				else
					nramcs(0) <= '1';
				end if;
			
				-- bank 1 ram, 512k-1M
				if (cpuaddrin_64k(23 downto 19) = "00001")
					then
					nramcs(1) <= '0';
				else
					nramcs(1) <= '1';
				end if;

				-- uppermost 512k, boot rom
				if (cpuaddrin_64k(23 downto 19) = "11111")
					then
					-- uppermost 512k, write only if allowed by prgrom
					nromcs <= '0';
					nromwe <= cpurnw or not(prgrom);
				elsif (islow64k = '1' 
						and cpurnw = '1'
						and bootrom = '1' 
						and isio = '0'
						and isvideo = '0'
						) then 
						-- low 64k, bootrom, read only, not I/O and not video
						nromcs <= '0';
						nromwe <= '1';
				else
					nromcs <= '1';
					nromwe <= '1';
				end if;
			end if;

		end process MemSel;
		
end Behavioral;

